#include <assert.h>
#include "Hierarchy.h"

Level LevelCreate(
    LevelRenderFunc render, 
    LevelExpandFunc expand,
    LevelBeginFunc begin,
    LevelDestroyFunc destroy)
{
  Level lev;
  lev.begin = begin;
  lev.render = render;
  lev.focus = 0x0;
  lev.expand = expand;
  lev.canFocus = 0x0;
  lev.canExpand = 0x0;
  lev.destroy = destroy;
  lev.focused = -1;
  lev.expanded = -1;
  lev.rendered = -1;
  lev.length = -1; // unknown length
  return lev;
}

void LevelBegin(Level *lev) {
  if (lev->begin)
    (*lev->begin)(lev);
  lev->rendered = -1;
}

int LevelRender(Level *lev, int index, FILE *outFile) {
  int res;
  if ((lev->length >= 0) && (index >= lev->length))
    return -1;
  if ((index <= lev->rendered) || (lev->rendered == -1))
    LevelBegin(lev);
  while (index != lev->rendered + 1) {
    res = (*lev->render)(lev, lev->rendered + 1, 0x0);
    if (res == -1) {
      lev->length = lev->rendered + 1;
      return -1;
    }
    lev->rendered = lev->rendered + 1;
  }
  res = (*lev->render)(lev, lev->rendered + 1, outFile);
  if (res != -1)
    lev->rendered = lev->rendered + 1;
  if ((lev->length < 0) && (-lev->length - 1 <= lev->rendered))
    lev->length = -(lev->rendered + 2);
  return res;
}

void LevelFocus(Level *lev, int index) {
  if (lev->focus)
    (*lev->focus)(lev, index);
  lev->focused = index;
}

Level LevelExpand(Level *lev, int index) {
  Level sublevel = (*lev->expand)(lev, index);
  lev->expanded = index;
  return sublevel;
}

int LevelCanFocus(Level *lev, int index) {
  if (lev->canFocus)
    return (*lev->canFocus)(lev, index);
  else
    return 1;
}

int LevelCanExpand(Level *lev, int index) {
  if (lev->canExpand)
    return (*lev->canExpand)(lev, index);
  else if (lev->expand)
    return 1;
  else
    return 0;
}

int LevelGetFocused(Level *lev) {
  return lev->focused;
}

int LevelGetExpanded(Level *lev) {
  return lev->expanded;
}

int LevelExists(Level *lev, int index) {
  if (lev->length >= 0)
    return index < lev->length;
  int len = -lev->length - 1; /* last KNOWN length */
  if (index < len)
    return 1;
  /* Now we need to iterate to discover if index is the end */
  int ret = LevelRender(lev, index, 0x0);
  return ret != -1;
}

void LevelDestroy(Level *lev) {
  if (lev->destroy)
    (*lev->destroy)(lev);
}

void LevelSetPosition(Level *lev, int index) {
  if (index == 0) {
    LevelBegin(lev);
    return;
  }
  LevelRender(lev, index - 1, 0x0);
}

/* ** Hier Body ** */
inline HierItem HierItemCreate(int level, int item) {
  HierItem hi;
  hi.level = level;
  hi.item = item;
  return hi;
}

void HierInit(Hier *hier, Level *lev) {
  if (lev) {
    hier->depth = 1;
    hier->levels[0] = *lev;
  } else {
    hier->depth = 0;
  }
  hier->focused = HierItemCreate(-1, -1);
  hier->offset  = HierItemCreate(-1, -1);
}

void HierFocus(Hier *hier, HierItem item) {
  hier->focused = item;
}

inline HierItem HierFirst(Hier *hier) {
  if (hier->depth == 0)
    return HierItemCreate(-1, -1);
  else if (!LevelExists(&hier->levels[0], 0))
    return HierItemCreate(-1, -1);
  else
    return HierItemCreate(0, 0);
}

HierItem HierNext(Hier *hier, HierItem item) {
  if (LevelExists(&hier->levels[item.level], item.item + 1)) {
    return HierItemCreate(item.level, item.item + 1);
  } else {
    return HierNext(hier, HierParent(hier, item)); 
  }
}

HierItem HierPrev(Hier *hier, HierItem item) {
  if (item.item == 0)
    return HierParent(hier, item);
  return HierItemCreate(item.level, item.item - 1);
}

HierItem HierParent(Hier *hier, HierItem item) {
  if (item.level == 0) {
    return HierItemCreate(-1, -1);
  } else {
    int lev = item.level - 1;
    return HierItemCreate( lev, LevelGetExpanded(&hier->levels[lev]));
  }
}

inline void HierFocusNext(Hier *hier) {
  HierFocus(hier, HierNext( hier, HierGetFocused(hier) ) );
}

inline void HierFocusPrev(Hier *hier) {
  HierFocus(hier, HierPrev( hier, HierGetFocused(hier) ) );
}

inline HierItem HierGetFocused(Hier *hier) {
  return hier->focused;
}

inline int HierGetDepth(Hier *hier) {
  return hier->depth;
}

inline HierItem HierGetOffset(Hier *hier) {
  return hier->offset;
}

inline void HierSetOffset(Hier *hier, HierItem offset) {
  hier->offset = offset;
}

void HierExpand(Hier *hier, HierItem item) {
  assert(hier->depth > 0);
  assert(item.level != -1);
  assert(item.item != -1);
  assert(item.level < HIER_MAX_DEPTH-1);
  HierUnexpand(hier, item.level);
  hier->levels[item.level + 1] = LevelExpand(&hier->levels[item.level], item.item);
  hier->depth++;
}

inline void HierUnexpand(Hier *hier, int level) {
  assert(level >= 0);
  while (hier->offset.level > level)
    hier->offset = HierParent(hier, hier->offset);
  while (hier->focused.level > level)
    hier->focused = HierParent(hier, hier->focused);
  if (level < hier->depth)
    hier->depth = level + 1;
}

