#ifndef HIERARCHY_H
#define HIERARCHY_H

#include <stdio.h>

/* *** Level *** */

#define LEVEL_DATA_SIZE 100

struct Level_;
typedef void (*LevelBeginFunc)(struct Level_ *lev);
typedef int (*LevelRenderFunc)(struct Level_ *lev, int index, FILE *outfile);
typedef void (*LevelFocusFunc)(struct Level_ *lev, int index);
typedef struct Level_ (*LevelExpandFunc)(struct Level_ *lev, int index);
typedef int (*LevelCanFocusFunc)(struct Level_ *lev, int index);
typedef int (*LevelCanExpandFunc)(struct Level_ *lev, int index);
typedef void (*LevelDestroyFunc)(struct Level_ *lev);

/** Level is more of an abstract class than a struct.  It has multiple virtual
 * functions and worst-case storage space.  The end user
 * should not use the function pointers in this struct
 * directly, but instead use the Level*() functions, which
 * provide defaults. **/
typedef struct Level_ { 
  /** [Re]start the list of items to render **/
  //void (*begin)(struct Level_ *lev);
  LevelBeginFunc begin;
  /** Render the next item in the level.  Note that index is only passed for
   * possible optimization purposes - index will always incrementally increase
   * unless begin() is called **/
  //int (*render)(struct Level_ *lev, int index, FILE *outFile);
  LevelRenderFunc render;
  /** Callback for item focus/unfocus.  Optional.  Called before
   * lev->focused is set so the old selection can be determined if required.
   * **/
  //void (*focus)(struct Level_ *lev, int index);
  LevelFocusFunc focus;
  //struct Level_ (*expand)(struct Level_ *lev, int index);
  LevelExpandFunc expand;
  //int (*canFocus)(struct Level_ *lev, int index);
  //int (*canExpand)(struct Level_ *lev, int index);
  LevelCanFocusFunc canFocus;
  LevelCanExpandFunc canExpand;
  LevelDestroyFunc destroy;
  int focused; /**< Currently focused item, or -1 for none **/
  int expanded; /**< Currently expanded item, or -1 for none **/
  int rendered; /**< Last rendered item, or -1 for none **/
  int length; /**< if negative then abs(length)-1 = the maximum KNOWN
                length.  if positive, then this is the definitive length **/
  char data[LEVEL_DATA_SIZE]; 
} Level;

/** Default initializer for Level **/
Level LevelCreate(
    LevelRenderFunc render, 
    LevelExpandFunc expand,
    LevelBeginFunc begin,
    LevelDestroyFunc destroy);

/** [Re]start the list of items to render **/
void LevelBegin(Level *lev);

/** Render an item within a Level 
 *
 * outFile - the specified output file descriptor.  If NULL,
 * then nothing is output.
 *
 * Returns the number of visible
 * characters actually output, or -1 if index is beyond the
 * end of available items.
 * **/
int LevelRender(Level *lev, int index, FILE *outFile);

/** "Focus" the specified item.  **/
void LevelFocus(Level *lev, int index);

/** "Expand" the specified item.  This construct and returns a
 * new Level structure containing the expanded sub-Level. **/
Level LevelExpand(Level *lev, int index);

/** Is specified item capable of being "focused"? **/
int LevelCanFocus(Level *lev, int index);

/** Is specified item capable of being "expanded"? **/
int LevelCanExpand(Level *lev, int index);

/** What is the index of the current focus? **/
int LevelGetFocused(Level *lev);

/** What is the index of the current expansion? **/
int LevelGetExpanded(Level *lev);

/** Boolean, returns true if the specified index exists within the level. **/
int LevelExists(Level *lev, int index);

void LevelDestroy(Level *lev);

/** Set level's position so that next call to LevelRender renders index -
 * ie, to position before index **/
void LevelSetPosition(Level *lev, int index);

/* *** Hierarchy Model *** */
#define HIER_MAX_DEPTH 5

/** Specifies a particular level and a particular item within that level. **/
typedef struct HierItem_ {
  int level;
  int item;
} HierItem;

HierItem HierItemCreate(int level, int item);

typedef struct Hier_ {
  Level levels[HIER_MAX_DEPTH];
  int depth; /**< Number of levels expanded == number of items in levels[] array **/
  HierItem focused; /**< Currently focused item **/
  HierItem offset; /**< First level/item to display on screen **/
} Hier;

void HierInit(Hier *hier, Level *level);

/** Set a new focused item within the currently expanded set of levels **/
void HierFocus(Hier *hier, HierItem item);

/** Get first, top level item in hierarchy **/
HierItem HierFirst(Hier *hier);

/** Return the next item in a logical traversal of the expanded tree **/
HierItem HierNext(Hier *hier, HierItem item);

/** Return the previous item in a logical traversal of the expanded tree **/
HierItem HierPrev(Hier *hier, HierItem item);

/** Return the parent item of the specified item
 *  ie, the expanded item in the level above item's level **/
HierItem HierParent(Hier *hier, HierItem item);

/** Select the next logical item to focus.   
 *  Equivalent to HierFocus(hier, HierNext(hier, HierGetFocused(hier))) **/
void HierFocusNext(Hier *hier);

/** Select the previous logical item to focus **/
void HierFocusPrev(Hier *hier);

/** Return currently focused item and level indicies **/
HierItem HierGetFocused(Hier *hier);

int HierGetDepth(Hier *hier);

HierItem HierGetOffset(Hier *hier);

void HierSetOffset(Hier *hier, HierItem offset);

/** Expand the specified item - automatically unexpand any expanded non-ancestors**/
void HierExpand(Hier *hier, HierItem item);

/** Contract to the specified level **/
void HierUnexpand(Hier *hier, int level);

#endif // HIERARCHY_H
