#include "WikiText.h"
#include <string.h>
#include <Platform.h>

const WToken tokens[] = {
  {"[[", 2, WT_OPEN,  WT_INT_LINK},
  {"]]", 2, WT_CLOSE, WT_INT_LINK},
  {"[", 1, WT_OPEN,  WT_EXT_LINK},
  {"]", 1, WT_CLOSE, WT_EXT_LINK},
  {"\\n",2, WT_FLAT,  WT_NEWLINE}, 
  {"{{", 2, WT_OPEN,  WT_TEMPLATE},
  {"}}", 2, WT_CLOSE, WT_TEMPLATE},
  {"|",  1, WT_FLAT,  WT_PIPE},
  {"'''''",5, WT_MATCH, WT_BOLD_ITALIC},
  {"'''",3, WT_MATCH, WT_BOLD},
  {"''", 2, WT_MATCH, WT_ITALIC},
  {"<",  1, WT_OPEN,  WT_HTML_TAG},
  {">",  1, WT_CLOSE, WT_HTML_TAG},
  {"=====",5, WT_MATCH,  WT_HEAD5},
  {"====",4, WT_MATCH,  WT_HEAD4},
  {"===",3, WT_MATCH,  WT_HEAD3},
  {"==", 2, WT_MATCH,  WT_HEAD2},
  {"=",  1, WT_FLAT,  WT_EQUALS}
};
#define TokensLen (sizeof(tokens)/sizeof(WToken))

/** Return the next "flat" token.
 *  Example tokens: '[[', ']]', '\n', '{{', '}}'. '|', '', ''', '<', '>'
 *  We do NOT assume that all tokens are meaninful, we just want to divide the text into
 *  all POSSIBLE tokens and can sort out the context later. **/
WToken WikiTextNextToken(char *text, int textLen) {
  int tokenLen=0;
  char *curText = text;
  while (textLen && (tokenLen < 256)) {
    for (int i=0; i<TokensLen; ++i) {
      int minlen = textLen < tokens[i].strLen ? textLen : tokens[i].strLen;
      if (!strncmp(tokens[i].str, curText, minlen)) {
        if (tokenLen) {
          WToken wt = {text, tokenLen, WT_FLAT, WT_TEXT};
          return wt; // new token is "next"
        } else  {
          WToken wt = tokens[i];
          wt.str = curText;
          return wt; // new token IS this token
        }
      }
    }
    ++tokenLen;
    --textLen;
    ++curText;
  }
  WToken wt = {text, tokenLen, WT_FLAT, WT_TEXT};
  return wt;
}

int WTTokenize(char *text, int textLen, WToken *tokens, int tokensLen) {
  int tokenCount = 0;
  while (1) {
    if (tokensLen == 0) break;
    if (textLen == 0) break;
    *tokens = WikiTextNextToken(text, textLen);
    text += tokens->strLen;
    textLen -= tokens->strLen;
    ++tokens;
    --tokensLen;
    ++tokenCount;
  }
  return tokenCount;
}

WToken *WTFirstChild(WToken *tokens, int tokensLen) {
  if (tokensLen < 2) return 0x0; // not enough tokens
  if (tokens->type != WT_OPEN) return 0x0; // not parent
  return tokens + 1;
}

WToken *WTGetChild(WToken *tokens, int tokensLen, int n) {
  WToken *tokensEnd = tokens + tokensLen;
  tokens = WTFirstChild(tokens, tokensLen);
  if (!tokens) return 0x0;
  tokensLen = tokensEnd - tokens;
  for (int i=0; i<n; ++i) {
    if (tokensLen == 0) return 0x0;
    tokens = WTNextSibling(tokens, tokensLen);
    if (!tokens) return 0x0;
    if (tokens->type == WT_CLOSE) return 0x0;
    tokensLen = tokensEnd - tokens;
  }
  return tokens;
}

WToken *WTGetLastChild(WToken *tokens, int tokensLen) {
  WToken *tokensEnd = tokens + tokensLen;
  tokens = WTFirstChild(tokens, tokensLen);
  if (!tokens) return 0x0;
  tokensLen = tokensEnd - tokens;
  WToken *lastToken = tokens;
  while (1) {
    if (tokensLen == 0) return 0x0;
    tokens = WTNextSibling(tokens, tokensLen);
    if (!tokens) return lastToken;
    if (tokens->type == WT_CLOSE) return tokens;
    tokensLen = tokensEnd - tokens;
  }
}

/** Get next sibling of current token **/
WToken *WTNextSibling(WToken *tokens, int tokensLen) {
  WToken *lastChild = WTGetLastChild(tokens, tokensLen);
  if (lastChild) {
    tokensLen += 1 + lastChild - tokens;
    tokens = 1 + lastChild;
  }
  char *target = tokens->str + tokens->strLen;
  while (1) {
    ++tokens;
    --tokensLen;
    if (tokensLen == 0) 
      return 0x0;
    if (tokens->str >= target) 
      return tokens;
  }
}

int WTPrintDebug(WToken *tokens, int tokensLen, int level) {
  int tokenCount = 0;
  while (1) {
    if (tokensLen - tokenCount == 0) 
      return tokenCount + 1;
    for (int i=0; i<level; ++i)
      DebugOut('.');
    for (int i=0; i<tokens[tokenCount].strLen; ++i)
      DebugOut(tokens[tokenCount].str[i]);
    DebugOut('\n');
    if (tokens[tokenCount].type == WT_CLOSE)
      return tokenCount + 1;
    if (tokensLen - tokenCount < 2)
      return tokenCount + 1;
    if (tokens[tokenCount].type == WT_OPEN) //parent
      tokenCount += WTPrintDebug(&tokens[tokenCount + 1], tokensLen - tokenCount - 1, level + 1);
    ++tokenCount;
  }
}

#if 0
int WikiTextFirstSegment(char *text, int textLen, WSegment *segments, int segmentsSize) 
{
  if (segmentsSize < 2) return 0;
  WSegment *tok = segments;
  ++segments;
  --segmentsSize;
  int nSeg=1;
  int level=0;

  do {
    *tok = WikiTextNextToken(text, textLen);
    if (tok->type == WT_OPEN) {
      ++level;
      if (level > 1)
        WikiTextFirstSegment(text, textLen, segments, segmentsSize);
    }
  } while ((level > 0) && (textLen > 0) && (segmentsSize > 0))
  return nSeg;
}

WSegment WikiTextSegment(char *text, int textLen, WSegment *segments, int segmentsSize) {
  WSegment seg;

  while (1) {
    if (textLen == 0) return seg;
  }
  return seg;
}
#endif
