#include "Glyphs.h"
#include "Errors.h"
#include <assert.h>
#include <stdlib.h>

inline void CellInit(Cell *c) {
  memset(c, sizeof(uint8_t), 0x0);
  c->glyphIndex=0xff;
}

Cell cellMap[CELLMAP_WIDTH][CELLMAP_HEIGHT];
int cellMapXOffset=0;
int cellMapYOffset=0;
const uint8_t *glyphMap=0x0; // bit packed glyph array
uint8_t glyphMapSize=0; // number of GLYPHS in array (NOT bytes)
uint8_t glyphWidth=0;
uint8_t glyphHeight=0;

void RenderCell8bpp(Cell c, unsigned char *b, int x, int y) {
  if (c.glyphIndex == 0xff)  /* FIXME: WRONG IF WE INVERT! */
    return;
  assert(c.glyphIndex < glyphMapSize);
  assert((glyphWidth == 8) && "glyph widths other than 8 is not yet supported!\n");
  for (int j=0; j < glyphHeight; ++j) {
    unsigned char r = *(GetGlyph(c.glyphIndex) + j);
    for (int i=0; i < glyphWidth; ++i) {
      uint8_t cc = r & (0x1 << (7-i));
      if (cc) RenderBufferSet(b, x + i + 2*c.shiftX, y + j + 2*c.shiftY, 0xff);
    }
  }
}

void RenderRow(unsigned char *b, int row) {
  for (int i=0; i<CELLMAP_WIDTH; ++i) {
    RenderCell8bpp(*CellMapGet(i, row), b, i*CELL_X_ADVANCE, row*CELL_Y_ADVANCE);
  }
}

uint8_t Neighbors(int cx, int cy) {
  uint8_t n = 0x0;
  n |= (CellMapGet(cx - 1, cy)->glyphIndex == 0xff) << 0x0;
  n |= (CellMapGet(cx, cy - 1)->glyphIndex == 0xff) << 0x1;
  n |= (CellMapGet(cx - 1, cy - 1)->glyphIndex == 0xff) << 0x2;
  return n;
}

void PlaceCell(int cx, int cy, Cell cell, int recurse) {
  if (CellMapGet(cx, cy)->glyphIndex == 0xff) {
    CellMapSet(cx, cy, cell);
    return;
  }
  uint8_t n = Neighbors(cx, cy);
  if (n & 0x4) { // prefer the top-left corner
    cell.shiftX += CELL_X_ADVANCE>>1;
    cell.shiftY += CELL_Y_ADVANCE>>1;
    CellMapSet(cx - 1, cy - 1, cell);
  } else if (n & 0x1) { // left
    cell.shiftX += CELL_X_ADVANCE>>1;
    CellMapSet(cx - 1, cy, cell);
  } else if (n & 0x2) { // top
    cell.shiftY += CELL_Y_ADVANCE>>1;
    CellMapSet(cx, cy - 1, cell);
  } else { // no neighbors empty - bump someone
    if (recurse == 0)
      ERRORreturn("PlaceCell failed");
    // start with top-left - we know the least about him
    PlaceCell(cx-1, cy-1, *CellMapGet(cx-1, cy-1), recurse - 1);
    if (!IsError) {
      // Success!  Put ourself in his spot
      cell.shiftX += CELL_X_ADVANCE>>1;
      cell.shiftY += CELL_Y_ADVANCE>>1;
      CellMapSet(cx - 1, cy - 1, cell);
      return;
    }
    PlaceCell(cx-1, cy, *CellMapGet(cx-1, cy), recurse - 1);
    if (!IsError) {
      // Success!
      cell.shiftX += CELL_X_ADVANCE>>1;
      CellMapSet(cx - 1, cy, cell);
      return;
    } 
    // Final attempt - if it throws, we just fail
    PlaceCell(cx, cy - 1, *CellMapGet(cx, cy - 1), recurse - 1);
    if (IsError) return;
    cell.shiftY += CELL_Y_ADVANCE>>1;
    CellMapSet(cx, cy - 1, cell);
    return;
  }
}

/** Add a Glyph to the given x/y pixel location **/
void AddGlyph(int x, int y, uint8_t glyphIndex, uint8_t bright, uint8_t invert) {
  Cell c;
  int cx = x/CELL_X_ADVANCE;
  int cy = y/CELL_Y_ADVANCE;
  CellInit(&c);
  c.shiftX = (x % CELL_X_ADVANCE) >> 1;
  c.shiftY = (y % CELL_Y_ADVANCE) >> 1;
  c.bright = bright;
  c.invert = invert;
  c.glyphIndex = glyphIndex;

  assert(glyphIndex < glyphMapSize);
  //printf("AddGlyph(%i, %i) => (%i, %i, sx %i, sy %i)\n", x, y, cx, cy, c.shiftX, c.shiftY);
  PlaceCell(cx, cy, c, 1);
}

void SetGlyphMap(const uint8_t *glyphMap_, uint8_t glyphMapSize_, uint8_t glyphWidth_, uint8_t glyphHeight_) {
  glyphMap = glyphMap_;
  glyphMapSize = glyphMapSize_;
  glyphWidth = glyphWidth_;
  glyphHeight = glyphHeight_;
}

void ClearCells() {
  for (int i=0; i<CELLMAP_HEIGHT; ++i)
    for (int j=0; j<CELLMAP_WIDTH; ++j)
      CellInit(&cellMap[j][i]);
}

void DebugPrintGlyphs() {
#ifdef UNIX
  printf("     ");
  for (int i=0; i<CELLMAP_WIDTH; ++i)
    printf("%3i", i);
  printf("\n");
  for (int j=0; j<CELLMAP_HEIGHT; ++j) {
    printf("R%03i ", j);
    for (int i=0; i<CELLMAP_WIDTH; ++i) {
      if (cellMap[i][j].glyphIndex == 0xff)
        printf("__ ");
      else
        printf("%2X ", cellMap[i][j].glyphIndex);
    }
    printf("\n");
  }
#endif
}

inline Cell *CellMapGet(int x, int y) {
  return &cellMap[mod(x + cellMapXOffset, CELLMAP_WIDTH)][mod(y + cellMapYOffset, CELLMAP_HEIGHT)];
}

inline void CellMapSet(int x, int y, Cell c) {
  cellMap[mod(x + cellMapXOffset, CELLMAP_WIDTH)][mod(y + cellMapYOffset, CELLMAP_HEIGHT)] = c;
}

inline void RenderBufferSet(unsigned char *buf, int x, int y, uint8_t value) {
  buf[mod(x, RenderBufferWidth()) + mod(y, RenderBufferHeight())*RenderBufferWidth()] = value;
}

inline uint8_t RenderBufferGet(unsigned char *buf, int x, int y) {
  return buf[mod(x, RenderBufferWidth()) + mod(y, RenderBufferHeight())*RenderBufferWidth()];
}

inline unsigned char *CalcRenderBufferLinePos(unsigned char *buf, int y) {
  return buf + mod(y, RenderBufferHeight()) * RenderBufferWidth();
}
