#include <assert.h>
#include <string.h>
#include <stdio.h>
#ifdef AVR
#  define NO_STDIO
#endif
#ifndef NO_STDIO
#  include "bgUtil2.h"
#endif

#include "ar.h"
#include "crc32.h"
#include "bgCore.h"
#include "Platform.h"
#include "bgUtil.h"

/* This is the current block context for the bgtree db system.  You will have
 * to swap this if you are using more than one table stored in more than one
 * file (or memory area, etc) */
BlockContext *bgtreeContext;

/** User-supplied function loads block as addr, returning result in buffer.
 * buffer is a pre-allocated array of size blkSize **/
Node *LoadBlock(BlockAddrT addr) {
  Node *n = LockBlockInCache(&bgtreeContext->cache, addr);
  if (n) return n; // else we need to load it
  n = AllocBlockFromCache(&bgtreeContext->cache, addr);
  assert(n && "LoadBlock: No unlocked space left in cache!");
  // Get block from disk
  FileSeek(bgtreeContext->offset + addr);
  ERRORassert();
  FileRead((char *) n, BlockSize);
  assert(!IsError);
  return n; 
}

void UnloadBlock(BlockAddrT addr, Node *ptr) {
  UnlockBlockInCache(&bgtreeContext->cache, addr);
}

/** User supplied function to allocate a new block **/
BlockAddrT AllocBlock() { assert(0); } // read only

void SaveBlock(BlockAddrT addr, Node *buffer) { assert(0); } // read only

int ArReadHeaders(ArFileRecord *records, int maxRecords) {
  char buf[60];
  uint32_t position=0;
  uint32_t numRecords=0;
  /* Read global header */
  do {
    ++position; // eat first line
  } while (FileGetC(FileGetHandle()) != '\n');

  do {
    int sz = FileRead(buf, 1);
    if (IsError || !sz) { //end of archive
      ClearError();
      break;
    }
    if (buf[0] == '\n') // record terminator
      break;

    FileRead(buf+1, 59);
    if (IsError) {
      fprintf(stderr, "Premature end of archive during header read\n");
      return -4;
    }
    ArParseHeader(&records[numRecords++], buf, &position);
    fprintf(stderr,"%2u: sz=%10u pos=%10u name=%s\n", (unsigned int) numRecords, (unsigned int) records[numRecords-1].sz, (unsigned int) records[numRecords-1].position, records[numRecords-1].fname);
    FileSeek(position);
    if (IsError) {
      fprintf(stderr, "Attempted seek in archive failed\n");
      return -4;
    }
  } while (1); 
  return numRecords;
}

// diskfilename parameter defined in patfs-stdio
const char *diskfilename;

int main(int argc, char **argv) {
  if (argc < 3) {
    fprintf(stderr, "USAGE: %s search <wiki.ar> <ArticleName> [<fat.img>]\n", argv[0]);
    return -1;
  }

  diskfilename = argc > 4 ? argv[4] : "/home/braddock/expire/100k/fat.img";
  fprintf(stderr, "Attempting to open FAT fs %s\n", diskfilename);
  PlatformInit();
  ERRORassert();

  const int maxRecords = 8;
  ArFileRecord records[maxRecords];
  int numRecords=0;

  FileOpenRO(argv[2]);
  if (IsError) {
    fprintf(stderr, "Failed to open archive %s\n", argv[2]);
    return -1;
  }
  numRecords = ArReadHeaders(records, maxRecords);
 
  // Setup text table context
  int idx;
  for (idx=0; (idx < numRecords) && strcmp(records[idx].fname, "text-100k.bgt"); ++idx) ;
  assert(idx < numRecords);
  BlockContext textTable;
  textTable.offset = records[idx].position + 60;
  InitBlockCache(&textTable.cache);
  // Setup text BLOB
  for (idx=0; (idx < numRecords) && strcmp(records[idx].fname, "text-100k.bgb"); ++idx) ;
  assert(idx < numRecords);
  uint32_t textBlobOffset = records[idx].position + 60;

  // Setup page table context
  for (idx=0; (idx < numRecords) && strcmp(records[idx].fname, "page-100k.bgt"); ++idx) ;
  assert(idx < numRecords);
  BlockContext pageTable;
  pageTable.offset = records[idx].position + 60;
  InitBlockCache(&pageTable.cache);
  // Setup page BLOB
  for (idx=0; (idx < numRecords) && strcmp(records[idx].fname, "page-100k.bgb"); ++idx) ;
  assert(idx < numRecords);
  uint32_t pageBlobOffset = records[idx].position + 60;

  bgtreeContext = &pageTable;
 
  if (!strcmp(argv[1],"search")) { 
    /* Perform Search */
    KeyT key = CalcCRC(argv[3], strlen(argv[3]));
    BAT outParent;
    int outIndex;
    BAT found = FindInternal(0, key, &outParent, &outIndex);
    if (found == BAInvalid) {
      fprintf(stderr, "%x NOT FOUND!\n", (unsigned int) key);
    } else {
      char buf[32];
      uint32_t strSize;
      FileSeek(pageBlobOffset + found);
      FileRead((char *) &strSize, 4);
      int sz = strSize > 31 ? 31 : strSize;
      FileRead(buf, sz);
      ERRORassert();
      buf[sz] = 0x0;
      //SetBright(); Puts(buf); UnsetBright(); Puts("\n");
      fwrite(buf, 1, sz, stdout);
      fputc('\n', stdout);
      
      uint32_t textId;
      FileSeek(pageBlobOffset + found + strSize + 4);
      FileRead((char *) &textId, 4);
      ERRORassert();
      fprintf(stderr, "%x\t%x = %x = %s\n", (unsigned int) key, (unsigned int) found, (unsigned int) textId, buf);

      /* Now lookup article Text */
      bgtreeContext = &textTable;
      found = FindInternal(0, textId, &outParent, &outIndex);
      if (found == BAInvalid) {
        fprintf(stderr, "Article Text ID %x Not Found\n", (unsigned int) textId);
        return -5;
      }
      uint32_t textSize;
      FileSeek(textBlobOffset + found);
      FileRead((char *) &textSize, 4);
      ERRORassert();
      //int dataPerKey=512;
      for (int i=0; i<textSize; i+=1) {
        FileRead(buf, 1); // kinda inefficient
        ERRORassert();
        fputc(buf[0], stdout);
#if 000
        if ((dataPerKey-- <= 0) && (cursorY > TERM_HEIGHT-3)) {
          Puts("\n-=[ Press Any Key To Continue ]=-");
          KeyEvent ke;
          while (!ReadKey(&ke)) {}; // wait for keypress
          while (ReadKey(&ke)) {}; // clear event queue
          Put('\n');
          dataPerKey = 512;
        }
#endif
      }
      fputc('\n',stdout);
    }
  } else if (!strcmp(argv[1],"print-text")) {
    bgtreeContext = &textTable;
#ifndef NO_STDIO
    NodeVisitDFS(&PrintVisitor, 0);
#else
    fprintf(stderr, "Not compiled with print-text enabled\n");
#endif
  } else if (!strcmp(argv[1],"print-page")) {
    bgtreeContext = &pageTable;
#ifndef NO_STDIO
    NodeVisitDFS(&PrintVisitor, 0);
#else
    fprintf(stderr, "Not compiled with print-text enabled\n");
#endif
  } else { 
    fprintf(stderr, "ERROR: Unknown command %s\n",argv[1]);
    return -2;
  }

  FileClose();

  return 0;
}
