#include <assert.h>
#include <string.h>
#include <stdio.h>
#ifdef AVR
#  define NO_STDIO
#endif
#ifndef NO_STDIO
#  include "bgUtil2.h"
#endif

#include "crc32.h"
#include "bgApi.h"
#include "Platform.h"
#include "bgUtil.h"

/* This is the current block context for the bgtree db system.  You will have
 * to swap this if you are using more than one table stored in more than one
 * file (or memory area, etc) */
BlockContext *bgtreeContext;

struct WikiApp {
  BlockContext textTable;
  BlockContext pageTable;
  BgTree textTree;
  char *textTableFname;
  char *textBlobFname;
  BgTree pageTree;
  char *pageTableFname;
  char *pageBlobFname;
};

void WikiAppInit(struct WikiApp *wa) {
  wa->textTable.offset = 0;
  wa->pageTable.offset = 0;
  InitBlockCache(&wa->textTable.cache);
  InitBlockCache(&wa->pageTable.cache);

  wa->textTableFname = "text-100k.bgt";
  wa->textBlobFname  = "text-100k.bgb";
  wa->textTree.topNode = 0;
  wa->pageTableFname = "page-100k.bgt";
  wa->pageBlobFname  = "page-100k.bgb";
  wa->pageTree.topNode = 0;
}

unsigned int WikiAppTitleSearch(struct WikiApp *wa, const char *title) {
  /* Hash title string */
  KeyT key = CalcCRC(title, strlen(title));
  /* Open table file and make it the current context */
  FileOpenRO(wa->pageTableFname);
  ERRORassert();
  bgtreeContext = &wa->pageTable;
  /* Perform search for title hash */
  //BAT outParent;
  //int outIndex;
  //BAT found = FindInternal(0, key, &outParent, &outIndex);
  Cursor c = SelectConsecutiveKeys(wa->pageTree, key);
  BAT found = CursorNext(&c);
  if (found == BAInvalid) {
    DebugPrint("%x NOT FOUND!\n", key);
    return (unsigned int)-1;
  } else {
    char buf[32];
    uint32_t strSize;
    FileOpenRO(wa->pageBlobFname);
    FileSeek(found);
    FileRead((char *) &strSize, 4);
    int sz = strSize > 31 ? 31 : strSize;
    FileRead(buf, sz);
    ERRORassert();
    buf[sz] = 0x0;
    //SetBright(); Puts(buf); UnsetBright(); Puts("\n");
    fwrite(buf, 1, sz, stdout);
    fputc('\n', stdout);

    uint32_t textId;
    FileSeek(found + strSize + 4);
    FileRead((char *) &textId, 4);
    ERRORassert();
    DebugPrint("%x\t%x = %x = %s\n", key, found, textId, buf);

    /* Now lookup article Text */
    FileOpenRO(wa->textTableFname);
    ERRORassert();
    bgtreeContext = &wa->textTable;
    //found = FindInternal(0, textId, &outParent, &outIndex);
    c = SelectConsecutiveKeys(wa->textTree, textId);
    found = CursorNext(&c);
    if (found == BAInvalid) {
      DebugPrint("Article Text ID %x Not Found\n", textId);
      return (unsigned int) -1;
    }
    return found;
  }
}

/** User-supplied function loads block as addr, returning result in buffer.
 * buffer is a pre-allocated array of size blkSize **/
Node *LoadBlock(BlockAddrT addr) {
  Node *n = LockBlockInCache(&bgtreeContext->cache, addr);
  if (n) return n; // else we need to load it
  n = AllocBlockFromCache(&bgtreeContext->cache, addr);
  assert(n && "LoadBlock: No unlocked space left in cache!");
  // Get block from disk
  FileSeek(bgtreeContext->offset + addr);
  ERRORassert();
  FileRead((char *) n, BlockSize);
  assert(!IsError);
  return n; 
}

void UnloadBlock(BlockAddrT addr, Node *ptr) {
  UnlockBlockInCache(&bgtreeContext->cache, addr);
}

/** User supplied function to allocate a new block **/
BlockAddrT AllocBlock() { assert(0); } // read only

void SaveBlock(BlockAddrT addr, Node *buffer) { assert(0); } // read only

// diskfilename parameter defined in patfs-stdio
const char *diskfilename;

int main(int argc, char **argv) {
  if (argc < 3) {
    DebugPrint("USAGE: %s search <ArticleName> [<fat.img>]\n", argv[0]);
    return -1;
  }

  diskfilename = argc > 3 ? argv[3] : "/home/braddock/expire/100k/fat.img";
  fprintf(stderr, "Attempting to open FAT fs %s\n", diskfilename);
  PlatformInit();
  ERRORassert();

  struct WikiApp wa;
  WikiAppInit(&wa);

  if (!strcmp(argv[1],"search")) { 
    /* Perform Search */
      BAT found = WikiAppTitleSearch(&wa, argv[2]);

      FileOpenRO(wa.textBlobFname);
      uint32_t textSize;
      FileSeek(found);
      FileRead((char *) &textSize, 4);
      ERRORassert();
      //int dataPerKey=512;
      for (int i=0; i<textSize; i+=1) {
        char buf[32];
        FileRead(buf, 1); // kinda inefficient
        ERRORassert();
        fputc(buf[0], stdout);
#if 000
        if ((dataPerKey-- <= 0) && (cursorY > TERM_HEIGHT-3)) {
          Puts("\n-=[ Press Any Key To Continue ]=-");
          KeyEvent ke;
          while (!ReadKey(&ke)) {}; // wait for keypress
          while (ReadKey(&ke)) {}; // clear event queue
          Put('\n');
          dataPerKey = 512;
        }
#endif
      }
      fputc('\n',stdout);
  } else if (!strcmp(argv[1],"print-text")) {
    FileOpenRO("text-100k.bgt");
    ERRORassert();
    bgtreeContext = &wa.textTable;
#ifndef NO_STDIO
    NodeVisitDFS(&PrintVisitor, 0);
#else
    DebugPrint("Not compiled with print-text enabled\n");
#endif
  } else if (!strcmp(argv[1],"print-page")) {
    FileOpenRO("page-100k.bgt");
    ERRORassert();
    bgtreeContext = &wa.pageTable;
#ifndef NO_STDIO
    NodeVisitDFS(&PrintVisitor, 0);
#else
    DebugPrint("Not compiled with print-text enabled\n");
#endif
  } else { 
    DebugPrint("ERROR: Unknown command %s\n",argv[1]);
    return -2;
  }

  FileClose();

  return 0;
}
