#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "ar.h"
#include "crc32.h"
#include "bgCore.h"
#include "bgUtil.h"
#include "bgUtil2.h"
#include "Errors.h"
ErrorsDeclaration

/* This is the current block context for the bgtree db system.  You will have
 * to swap this if you are using more than one table stored in more than one
 * file (or memory area, etc) */
BlockContext *bgtreeContext;

/** User-supplied function loads block as addr, returning result in buffer.
 * buffer is a pre-allocated array of size blkSize **/
Node *LoadBlock(BlockAddrT addr) {
  Node *n = LockBlockInCache(&bgtreeContext->cache, addr);
  if (n) return n; // else we need to load it
  n = AllocBlockFromCache(&bgtreeContext->cache, addr);
  assert(n && "LoadBlock: No unlocked space left in cache!");
  // Get block from disk
  int res = fseek(bgtreeContext->f, bgtreeContext->offset + addr, SEEK_SET);
  assert(!res && "LoadBlock: fseek failed");
  res = fread(n, BlockSize, 1, bgtreeContext->f);
  assert(res == 1);
  return n; 
}

void UnloadBlock(BlockAddrT addr, Node *ptr) {
  UnlockBlockInCache(&bgtreeContext->cache, addr);
}

/** User supplied function to allocate a new block **/
BlockAddrT AllocBlock() { assert(0); } // read only

void SaveBlock(BlockAddrT addr, Node *buffer) { assert(0); } // read only

int ArReadHeaders(FILE *f, ArFileRecord *records, int maxRecords) {
  char buf[60];
  uint32_t position=0;
  uint32_t numRecords=0;
  /* Read global header */
  do {
    ++position; // eat first line
  } while (fgetc(f) != '\n');

  do {
    int n = fread(buf, 1, 1, f);
    if (n != 1) //end of archive
      break;
    if (buf[0] == '\n') // record terminator
      break;

    n = fread(buf+1, 1, 59, f);
    if (n < 59) {
      fprintf(stderr, "Premature end of archive during header read\n");
      return -4;
    }
    ArParseHeader(&records[numRecords++], buf, &position);
    fprintf(stderr, "%2i: sz=%10u pos=%10u name=%s\n", numRecords, records[numRecords-1].sz, records[numRecords-1].position, records[numRecords-1].fname);
    n = fseek(f, position, SEEK_SET);
    if (n != 0) {
      fprintf(stderr, "Attempted seek in archive failed\n");
      return -4;
    }
  } while (1); 
  return numRecords;
}

int main(int argc, char **argv) {
  if (argc < 3) {
    fprintf(stderr, "USAGE: %s search <wiki.ar> <ArticleName>\n", argv[0]);
    return -1;
  }
  const int maxRecords = 8;
  ArFileRecord records[maxRecords];
  int numRecords=0;

  FILE *f = fopen(argv[2],"rb");
  if (!f) {
    fprintf(stderr, "Failed to open archive %s\n", argv[2]);
    return -1;
  }
  numRecords = ArReadHeaders(f, records, maxRecords);
 
  // Setup text table context
  int idx;
  for (idx=0; (idx < numRecords) && strcmp(records[idx].fname, "text-100k.bgt"); ++idx) ;
  assert(idx < numRecords);
  BlockContext textTable;
  textTable.f = f;
  textTable.offset = records[idx].position + 60;
  InitBlockCache(&textTable.cache);
  // Setup text BLOB
  for (idx=0; (idx < numRecords) && strcmp(records[idx].fname, "text-100k.bgb"); ++idx) ;
  assert(idx < numRecords);
  uint32_t textBlobOffset = records[idx].position + 60;

  // Setup page table context
  for (idx=0; (idx < numRecords) && strcmp(records[idx].fname, "page-100k.bgt"); ++idx) ;
  assert(idx < numRecords);
  BlockContext pageTable;
  pageTable.f = f;
  pageTable.offset = records[idx].position + 60;
  InitBlockCache(&pageTable.cache);
  // Setup page BLOB
  for (idx=0; (idx < numRecords) && strcmp(records[idx].fname, "page-100k.bgb"); ++idx) ;
  assert(idx < numRecords);
  uint32_t pageBlobOffset = records[idx].position + 60;

  bgtreeContext = &pageTable;
 
  if (!strcmp(argv[1],"search")) { 
    /* Perform Search */
    KeyT key = CalcCRC(argv[3], strlen(argv[3]));
    BAT outParent;
    int outIndex;
    BAT found = FindInternal(0, key, &outParent, &outIndex);
    if (found == BAInvalid) {
      printf("%x NOT FOUND!\n", key);
    } else {
      char buf[32];
      int res;
      fseek(f, pageBlobOffset + found, SEEK_SET);
      res = UnpackStringFromFile(f, buf, 32);
      assert(res >= 0);
      buf[res-4] = 0x0; // in case of partial read
      uint32_t textId;
      fseek(f, pageBlobOffset + found + res, SEEK_SET);
      textId = UnpackUIntFromFile(f);
      ERRORassert();
      printf("%x\t%x = %x = %s\n", key, found, textId, buf);

      /* Now lookup article Text */
      bgtreeContext = &textTable;
      found = FindInternal(0, textId, &outParent, &outIndex);
      if (found == BAInvalid) {
        fprintf(stderr, "Article Text ID %x Not Found\n", textId);
        return -5;
      }
      uint32_t textSize;
      fseek(f, textBlobOffset + found , SEEK_SET);
      textSize = UnpackUIntFromFile(f);
      ERRORassert();
      for (int i=0; i<textSize; i+=1) {
        int n = fread(buf, 1, 1, f); // kinda inefficient
        if (!n) break;
        fwrite(buf, 1, n, stdout);
      }
      printf("\n");
    }
  } else if (!strcmp(argv[1],"print-text")) {
    bgtreeContext = &textTable;
    NodeVisitDFS(&PrintVisitor, 0);
  } else if (!strcmp(argv[1],"print-page")) {
    bgtreeContext = &pageTable;
    NodeVisitDFS(&PrintVisitor, 0);
  } else { 
    fprintf(stderr, "ERROR: Unknown command %s\n",argv[1]);
    return -2;
  }

  fclose(f);

  return 0;
}
